from dataclasses import dataclass


@dataclass
class NormalizerConfig:
    unify_level: str = "lexeme"
    taigen: bool = True
    yougen: bool = False
    expansion: str = "from_another"
    other_language: bool = True
    alias: bool = True
    old_name: bool = True
    misuse: bool = True
    alphabetic_abbreviation: bool = True
    non_alphabetic_abbreviation: bool = True
    alphabet: bool = True
    orthographic_variation: bool = True
    missspelling: bool = True
    custom_synonym: bool = True

    """
    NormalizerConfig class is a class that holds normalization settings.
    You can change the normalization settings by creating an instance of this class and passing it as an argument to the normalize method.
    Each field of this class holds a flag representing the normalization setting.

    Args:
        unify_level（default="lexeme"）: A flag specifying the unification level. The default "lexeme" unifies those with the same lexeme number. The "word_form" option unifies those with the same word form number. The "abbreviation" option unifies those with the same abbreviation number.
        taigen（default=True）: A flag indicating whether to include nouns in the unification. The default is to include. If you do not include, specify False.
        yougen（default=False）: A flag indicating whether to include verbs in the unification. The default is not to include. If you include, specify True. However, verbs are
        expansion（default="from_another"）: A control flag for synonym expansion. The default is to expand only those with an expansion control flag of 0. Specify "ANY" to expand those with an expansion control flag that is always expanded.
        other_language（default=True）: A flag indicating whether to normalize languages other than Japanese to Japanese. The default is to normalize. If you do not normalize, specify False.
        alias（default=True）: A flag indicating whether to normalize aliases. The default is to normalize. If you do not normalize, specify False.
        old_name（default=True）: A flag indicating whether to normalize old names. The default is to normalize. If you do not normalize, specify False.
        misuse（default=True）: A flag indicating whether to normalize misuse. The default is to normalize. If you do not normalize, specify False.
        alphabetic_abbreviation（default=True）: A flag indicating whether to normalize alphabetic abbreviations. The default is to normalize. If you do not normalize, specify False.
        non_alphabetic_abbreviation（default=True）: A flag indicating whether to normalize Japanese abbreviations. The default is to normalize. If you do not normalize, specify False.
        alphabet（default=True）: A flag indicating whether to normalize spelling inconsistencies in alphabets. The default is to normalize. If you do not normalize, specify False.
        orthographic_variation（default=True）: A flag indicating whether to normalize orthographic variations. The default is to normalize. If you do not normalize, specify False.
        missspelling（default=True）: A flag indicating whether to normalize misspellings. The default is to normalize. If you do not normalize, specify False.
        custom_synonym（default=True）: A flag indicating whether to use custom_synonyms set by the user. The default is to use. If you do not use, specify False.

    Examples:
        ```
        config = NormalizerConfig(taigen=0, yougen=1)
        normalizer = SynonymNormalizer(config)
        result = normalizer.normalize("テストを実行する")
        ```

    See Also:
        SudachiDictの同義語辞書ソース:
        https://github.com/WorksApplications/SudachiDict/blob/develop/docs/synonyms.md#%E5%90%8C%E7%BE%A9%E8%AA%9E%E8%BE%9E%E6%9B%B8%E3%82%BD%E3%83%BC%E3%82%B9-%E3%83%95%E3%82%A9%E3%83%BC%E3%83%9E%E3%83%83%E3%83%88


    日本語訳:
    NormalizerConfigクラスは正規化の設定を保持するクラスです。
    このクラスのインスタンスを生成し、normalizeメソッドに引数として渡すことで、正規化の設定を変更することができます。
    このクラスの各フィールドは、正規化の設定を表すフラグを保持します。

    Args:
        unify_level（default="lexeme"）:統一レベルを指定するフラグ。デフォルト"lexeme"はlexeme（語彙素）番号が同じもので統一。"word_form"オプションはwor_form（語形）番号が同じものでの統一。"abbreviation"オプションはabbreviation（略語）番号が同じものでの統一。
        taigen（default=True）: 統一するのに体言を含むかどうかのフラグ。デフォルトは含む。含まない場合はFalseを指定。
        yougen（default=False）: 統一するのに用言を含むかどうかのフラグ。デフォルトは含まない。含む場合はTrueを指定。ただし用言は
        expansion（default="from_another"）: 同義語展開の制御フラグ。デフォルトは同義語辞書の展開制御フラグが0のもののみ展開。"ANY"を指定すると展開制御フラグが常に展開する。
        other_language（default=True）: 日本語以外の言語を日本語に正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        alias（default=True）: 別称を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        old_name（default=True）: 旧称を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        misuse（default=True）: 誤用を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        alphabetic_abbreviation（default=True）: アルファベットの略語を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        non_alphabetic_abbreviation（default=True）: 日本語の略語を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        alphabet（default=True）: アルファベットの表記揺れを正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        orthographic_variation（default=True）: 異表記を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        missspelling（default=True）: 誤表記を正規化するかどうかのフラグ。デフォルトは正規化する。正規化しない場合はFalseを指定。
        custom_synonym（default=True）: ユーザーが設定したcustom_synonymを使用するかどうかのフラグ。デフォルトは使用する。使用しない場合はFalseを指定。

    """



import csv
import json
from collections import defaultdict
from typing import Dict, List, Optional, Set, Union

from yurenizer.entities import (
    AlphabeticAbbreviation,
    Alphabet,
    Expansion,
    FlgInput,
    NonAlphabeticAbbreviation,
    Missspelling,
    CusotomSynonym,
    OrthographicVariation,
    UnifyLevel,
    Taigen,
    Yougen,
    OtherLanguage,
    Alias,
    OldName,
    Misuse,
    TaigenOrYougen,
    FlgExpantion,
    WordForm,
    Abbreviation,
    SpellingInconsistency,
    SynonymField,
    Synonym,
    SudachiDictType,
    NormalizerConfig,
)
from sudachipy import Morpheme, dictionary, tokenizer


class SynonymNormalizer:
    def __init__(
        self,
        synonym_file_path: str,
        sudachi_dict: SudachiDictType = SudachiDictType.FULL.value,
        custom_synonyms_file: Optional[str] = None,
    ) -> None:
        """
        Initialize the tool for unifying spelling variations using SudachiDict's synonym dictionary
        （SudachiDictの同義語辞書を使用した表記ゆれ統一ツールの初期化）

        Args:
            synonym_file_path: Path to the SudachiDict synonym file（SudachiDictの同義語ファイルへのパス）
            sudachi_dict: SudachiDict type (default: full)（SudachiDictのタイプ（デフォルト: full））
            custom_synonyms_file: Path to the custom synonym definition file（カスタム同義語定義ファイルへのパス）

        Example:
            normalizer = SynonymNormalizer(synonym_file_path="synonyms.txt")
        """
        # Initialize Sudachi
        sudachi_dic = dictionary.Dictionary(dict=sudachi_dict)
        self.tokenizer_obj = sudachi_dic.create()
        self.mode = tokenizer.Tokenizer.SplitMode.C

        # Part-of-speech matching
        self.taigen_matcher = sudachi_dic.pos_matcher(lambda x: x[0] == "名詞")
        self.yougen_matcher = sudachi_dic.pos_matcher(lambda x: x[0] in ["動詞", "形容詞"])

        # Load synonyms from SudachiDict's synonym file
        synonyms = self.load_sudachi_synonyms(synonym_file_path)
        self.synonyms = {int(k): v for k, v in synonyms.items()}

        # Load custom synonyms
        self.custom_synonyms = {}
        if custom_synonyms_file:
            self.custom_synonyms = self.load_custom_synonyms(custom_synonyms_file)

    def load_sudachi_synonyms(self, synonym_file: str) -> Dict[int, List[Synonym]]:
        """
        Load synonym information from SudachiDict's synonyms.txt（SudachiDictのsynonyms.txtから同義語情報を読み込む）

        Args:
            synonym_file: Path to the SudachiDict synonym file（SudachiDictの同義語ファイルへのパス）

        Returns:
            Synonym information dictionary（同義語情報の辞書）

        Example:
            synonyms = load_sudachi_synonyms("synonyms.txt")
        """
        with open(synonym_file, "r", encoding="utf-8") as f:
            reader = csv.reader(f)
            data = [row for row in reader]

        synonyms = defaultdict(list)
        for line in data:
            if not line:
                continue
            synonyms[line[0]].append(
                Synonym(
                    taigen_or_yougen=int(line[1]),  # Taigen or Yougen
                    flg_expansion=int(line[2]),  # Expansion control flag
                    lexeme_id=int(line[3].split("/")[0]),  # Lexeme number within the group
                    word_form=int(line[4]),  # Word form type within the same lexeme
                    abbreviation=int(line[5]),  # Abbreviation
                    spelling_inconsistency=int(line[6]),  # Spelling inconsistency information
                    field=line[7],  # Field information
                    lemma=line[8],  # Lemma
                )
            )
        return synonyms

    def load_custom_synonyms(self, file_path: str) -> Dict[str, Set[str]]:
        """
        Load custom synonym definition JSON file（カスタム同義語定義JSONファイルを読み込む）

        Args:
            file_path: Path to the custom synonym definition JSON file（カスタム同義語定義JSONファイルへのパス）

        JSON format（JSONフォーマット）:
        {
            "standard_form": ["synonym1", "synonym2", ...],
            ...
        }
        Returns:
            Custom synonym definition（カスタム同義語定義）

        Example:
            custom_synonyms = load_custom_synonyms("custom_synonyms.json")
        """
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                custom_synonyms = json.load(f)
                custom_synonyms = {k: set(v) for k, v in custom_synonyms.items() if v}
                return custom_synonyms

        except Exception as e:
            raise ValueError(f"カスタム同義語定義の読み込みに失敗しました: {e}")

    def get_standard_yougen(self, morpheme: Morpheme, expansion: Expansion) -> str:
        """
        Get the standard representation of a verb or adjective（動詞または形容詞の標準形を取得）

        Args:
            morpheme: Morpheme information（形態素情報）
            expansion: Synonym expansion control flag（同義語展開の制御フラグ）

        Returns:
            Standard form（同義語が見つからない場合は元の単語）

        Example:
            standard_yougen = get_standard_yougen(morpheme, expansion)
        """

        # Search for synonym group from the group dictionary
        synonym_group = self.get_synonym_group(morpheme)
        if synonym_group:
            yougen_group = [s for s in synonym_group if s.taigen_or_yougen == TaigenOrYougen.YOUGEN.value]
            if expansion == Expansion.ANY:
                return yougen_group[0].lemma
            elif expansion == Expansion.FROM_ANOTHER:
                if {y.lemma: y.flg_expansion for y in yougen_group}.get(morpheme.normalized_form()) == 0:
                    return yougen_group[0].lemma
        return morpheme.surface()

    def get_standard_taigen(self, morpheme: Morpheme, expansion: Expansion) -> str:
        """
        Get the standard representation of a noun（名詞の標準形を取得）

        Args:
            morpheme: Morpheme information（形態素情報）
            expansion: Synonym expansion control flag（同義語展開の制御フラグ）

        Returns:
            Standard form（同義語が見つからない場合は元の単語）

        Example:
            standard_taigen = get_standard_taigen(morpheme, expansion)
        """
        # Search for synonym group from the group dictionary
        synonym_group = self.get_synonym_group(morpheme)
        if synonym_group:
            if expansion == Expansion.ANY:
                return synonym_group[0].lemma
            elif expansion == Expansion.FROM_ANOTHER:
                if {s.lemma: s.flg_expansion for s in synonym_group}.get(morpheme.normalized_form()) == 0:
                    return synonym_group[0].lemma
        return morpheme.surface()

    def get_custom_synonym(self, morpheme: Morpheme) -> Optional[str]:
        """
        Get the custom synonym representation of a word（単語のカスタム同義語表現を取得）

        Args:
            morpheme: Morpheme information（形態素情報）

        Returns:
            Custom synonym representation（カスタム同義語表現）

        Example:
            custom_representation = get_custom_synonym(morpheme)
        """
        custom_representation = self.normalize_word_by_custom_synonyms(morpheme.surface())
        if custom_representation:
            return custom_representation
        return None

    def normalize(
        self,
        text: str,
        config: NormalizerConfig = NormalizerConfig(),
    ) -> str:
        """
        Normalize text by unifying spelling variations and synonyms（表記ゆれと同義語を統一してテキストを正規化する）

        Args:
            text: Text to normalize（正規化するテキスト）
            config: Normalization options（正規化オプション）

        Returns:
            Normalized text（正規化されたテキスト）

        Example:
            normalized_text = normalize(text)
        """
        if not text:
            raise ValueError("テキストが空です")
        flg_input = FlgInput(
            unify_level=UnifyLevel.from_str(config.unify_level),
            taigen=Taigen.from_int(config.taigen),
            yougen=Yougen.from_int(config.yougen),
            expansion=Expansion.from_str(config.expansion),
            other_language=OtherLanguage.from_int(config.other_language),
            alias=Alias.from_int(config.alias),
            old_name=OldName.from_int(config.old_name),
            misuse=Misuse.from_int(config.misuse),
            alphabetic_abbreviation=AlphabeticAbbreviation.from_int(config.alphabetic_abbreviation),
            non_alphabetic_abbreviation=NonAlphabeticAbbreviation.from_int(config.non_alphabetic_abbreviation),
            alphabet=Alphabet.from_int(config.alphabet),
            orthographic_variation=OrthographicVariation.from_int(config.orthographic_variation),
            missspelling=Missspelling.from_int(config.missspelling),
            custom_synonym=CusotomSynonym.from_int(config.custom_synonym),
        )
        # Since the conditions are hierarchical, if the lower-level condition is true, the upper-level condition is also set to true
        if (
            flg_input.alphabet == Alphabet.ENABLE
            or flg_input.orthographic_variation == OrthographicVariation.ENABLE
            or flg_input.missspelling == Missspelling.ENABLE
        ):
            flg_input.alphabetic_abbreviation = AlphabeticAbbreviation.ENABLE
            flg_input.non_alphabetic_abbreviation = NonAlphabeticAbbreviation.ENABLE

        if (
            flg_input.alphabetic_abbreviation == AlphabeticAbbreviation.ENABLE
            or flg_input.non_alphabetic_abbreviation == NonAlphabeticAbbreviation.ENABLE
        ):
            flg_input.other_language = OtherLanguage.ENABLE
            flg_input.alias = Alias.ENABLE
            flg_input.old_name = OldName.ENABLE
            flg_input.misuse = Misuse.ENABLE
        return self.__normalize_text(text=text, flg_input=flg_input)

    def __normalize_text(self, text: str, flg_input: FlgInput) -> str:
        """
        Normalize text by unifying spelling variations and synonyms（表記ゆれと同義語を統一してテキストを正規化する）

        Args:
            text: Text to normalize（正規化するテキスト）
            flg_input: Normalization options（正規化オプション）

        Returns:
            Normalized text（正規化されたテキスト）

        Example:
            normalized_text = __normalize_text(text, flg_input)
        """
        morphemes = self.get_morphemes(text)
        normalized_parts = []
        for morpheme in morphemes:
            normalized_parts.append(self.normalize_word(morpheme, flg_input))
        return "".join(normalized_parts)

    def normalize_word(self, morpheme: Morpheme, flg_input: FlgInput) -> str:
        """
        Normalize a word by unifying spelling variations and synonyms（表記ゆれと同義語を統一して単語を正規化する）

        Args:
            morpheme: Morpheme information（形態素情報）
            flg_input: Normalization options（正規化オプション）

        Returns:
            Normalized word（正規化された単語）

        Example:
            normalized_word = normalize_word(morpheme, flg_input)
        """
        # Use custom synonym definitions
        custom_representation = self.get_custom_synonym(morpheme)
        if custom_representation:
            return custom_representation
        # Determine whether it's yougen or taigen
        is_yougen = self.yougen_matcher(morpheme)
        is_taigen = self.taigen_matcher(morpheme)
        if (flg_input.yougen == Yougen.INCLUDE and is_yougen) or (flg_input.taigen == Taigen.INCLUDE and is_taigen):
            pass
        else:
            return morpheme.surface()

        # Get synonym group for each taigen or yougen
        synonym_group = self.get_synonym_group(morpheme, is_yougen, is_taigen)
        if not synonym_group:
            return morpheme.surface()

        # Change subsequent processing according to the expansion control flag
        if flg_input.expansion == Expansion.ANY:
            if not self.is_input_word_expansion_any_or_from_another(morpheme, synonym_group):
                return morpheme.surface()
        elif flg_input.expansion == Expansion.FROM_ANOTHER:
            if not self.is_input_word_expansion_from_another(morpheme, synonym_group):
                return morpheme.surface()
        else:
            return morpheme.surface()

        # Narrow down the synonym group to the same lexeme
        if (
            flg_input.other_language == OtherLanguage.ENABLE
            or flg_input.alias == Alias.ENABLE
            or flg_input.old_name == OldName.ENABLE
            or flg_input.misuse == Misuse.ENABLE
        ):
            synonym_group = self.get_represent_synonym_group_lexeme_id(flg_input, morpheme, synonym_group)
        if not synonym_group:
            return morpheme.surface()

        # Get the synonym group with the same word form
        if (
            flg_input.alphabetic_abbreviation == AlphabeticAbbreviation.ENABLE
            or flg_input.non_alphabetic_abbreviation == NonAlphabeticAbbreviation.ENABLE
            or flg_input.alphabet == Alphabet.ENABLE
            or flg_input.orthographic_variation == OrthographicVariation.ENABLE
            or flg_input.missspelling == Missspelling.ENABLE
        ):
            synonym_group = self.get_represent_synonym_group_by_same_word_form(flg_input, morpheme, synonym_group)

        if not synonym_group:
            return morpheme.surface()

        # Get the synonym group with the same abbreviation
        if (
            flg_input.alphabet == Alphabet.ENABLE
            or flg_input.orthographic_variation == OrthographicVariation.ENABLE
            or flg_input.missspelling == Missspelling.ENABLE
        ):
            synonym_group = self.get_represent_synonym_group_by_same_abbreviation(flg_input, morpheme, synonym_group)
        if not synonym_group:
            return morpheme.surface()

        # Obtain the representative notation from the synonym group. Narrow down according to the expansion control flag
        represent_synonym = synonym_group[0]
        if represent_synonym:
            return represent_synonym.lemma
        return morpheme.surface()

    def is_input_word_expansion_any_or_from_another(self, morpheme: Morpheme, synonym_group: List[Synonym]) -> bool:
        """
        Judge whether the input word is expanded by synonyms（入力単語が同義語展開されるかどうかを判断する）

        Args:
            morpheme: Morpheme information（形態素情報）
            synonym_group: Synonym group（同義語グループ）

        Returns:
            True if the input word is expanded by synonyms, False otherwise（同義語展開される場合はTrue, されない場合はFalse）

        Example:
            is_expansion = is_input_word_expansion_any_or_from_another(morpheme, synonym_group)
        """
        flg_expansion = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.FLG_EXPANSION)
        if flg_expansion in (FlgExpantion.ANY.value, FlgExpantion.FROM_ANOTHER.value):
            return True
        return False

    def is_input_word_expansion_from_another(self, morpheme: Morpheme, synonym_group: List[Synonym]) -> bool:
        """
        Judge whether the input word is expanded by synonyms from another（入力単語が他の同義語展開されるかどうかを判断する）

        Args:
            morpheme: Morpheme information（形態素情報）
            synonym_group: Synonym group（同義語グループ）

        Returns:
            True if the input word is expanded by synonyms from another, False otherwise（他の同義語展開される場合はTrue, されない場合はFalse）

        Example:
            is_expansion = is_input_word_expansion_from_another(morpheme, synonym_group)
        """
        flg_expansion = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.FLG_EXPANSION)
        if flg_expansion == FlgExpantion.ANY.value:
            return True
        return False

    def get_represent_synonym_group_lexeme_id(
        self, flg_input: FlgInput, morpheme: Morpheme, synonym_group: List[Synonym]
    ) -> List[Synonym]:
        """
        Get the synonym group of the same lexeme id（同じ語彙素IDの同義語グループを取得）

        Args:
            morpheme: Morpheme information（形態素情報）
            synonym_group: Synonym group（同義語グループ）

        Returns:
            Synonym object list（Synonymオブジェクトのリスト）

        Example:
            synonym_group = get_represent_synonym_group_lexeme_id(flg_input, morpheme, synonym_group)
        """
        is_expansion = False
        filtered_synonym_group = []
        lexeme_id = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.LEXEME_ID)
        word_form = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.WORD_FORM)
        if word_form == WordForm.REPRESENTATIVE.value:
            is_expansion = True
        elif flg_input.other_language == OtherLanguage.ENABLE and word_form == WordForm.TRANSLATION.value:
            is_expansion = True
        elif flg_input.alias == Alias.ENABLE and word_form == WordForm.ALIAS.value:
            is_expansion = True
        elif flg_input.old_name == OldName.ENABLE and word_form == WordForm.OLD_NAME.value:
            is_expansion = True
        elif flg_input.misuse == Misuse.ENABLE and word_form == WordForm.MISUSE.value:
            is_expansion = True
        if is_expansion:
            filtered_synonym_group = [s for s in synonym_group if s.lexeme_id == lexeme_id]
        return filtered_synonym_group

    def get_represent_synonym_group_by_same_word_form(
        self, flg_input: FlgInput, morpheme: Morpheme, synonym_group: List[Synonym]
    ) -> List[Synonym]:
        """
        Get the synonym group of the same word form（同じ語形の同義語グループを取得）

        Args:
            morpheme: Morpheme information（形態素情報）
            synonym_group: Synonym group（同義語グループ）

        Returns:
            Synonym object list（Synonymオブジェクトのリスト）

        Example:
            synonym_group = get_represent_synonym_group_by_same_word_form(flg_input, morpheme, synonym_group)
        """
        is_expansion = False
        filtered_synonym_group = []
        word_form = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.WORD_FORM)
        abbreviation = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.ABBREVIATION)
        if abbreviation == Abbreviation.REPRESENTATIVE.value:
            is_expansion = True
        elif (
            flg_input.alphabetic_abbreviation == AlphabeticAbbreviation.ENABLE
            and abbreviation == Abbreviation.ALPHABET.value
        ):
            is_expansion = True
        elif (
            flg_input.non_alphabetic_abbreviation == NonAlphabeticAbbreviation.ENABLE
            and abbreviation == Abbreviation.NOT_ALPHABET.value
        ):
            is_expansion = True
        if is_expansion:
            if flg_input.unify_level == UnifyLevel.LEXEME:
                return synonym_group
            elif flg_input.unify_level in (UnifyLevel.WORD_FORM, UnifyLevel.ABBREVIATION):
                filtered_synonym_group = [s for s in synonym_group if s.word_form == word_form]
        return filtered_synonym_group

    def get_represent_synonym_group_by_same_abbreviation(
        self, flg_input: FlgInput, morpheme: Morpheme, synonym_group: List[Synonym]
    ) -> List[Synonym]:
        """
        Get the synonym group of the same abbreviation（同じ略語・略称の同義語グループを取得）

        Args:
            morpheme: Morpheme information（形態素情報）
            synonym_group: Synonym group（同義語グループ）

        Returns:
            Synonym object list（Synonymオブジェクトのリスト）

        Example:
            synonym_group = get_represent_synonym_group_by_same_abbreviation(flg_input, morpheme, synonym_group)
        """
        is_expansion = False
        filtered_synonym_group = []
        abbreviation = self.get_synonym_value_from_morpheme(morpheme, synonym_group, SynonymField.ABBREVIATION)
        spelling_inconsistency = self.get_synonym_value_from_morpheme(
            morpheme, synonym_group, SynonymField.SPELLING_INCONSISTENCY
        )
        if spelling_inconsistency == SpellingInconsistency.REPRESENTATIVE.value:
            is_expansion = True
        elif flg_input.alphabet == Alphabet.ENABLE and spelling_inconsistency == SpellingInconsistency.ALPHABET.value:
            is_expansion = True
        elif (
            flg_input.orthographic_variation == OrthographicVariation.ENABLE
            and spelling_inconsistency == SpellingInconsistency.ORTHOGRAPHIC_VARIATION.value
        ):
            is_expansion = True
        elif (
            flg_input.missspelling == Missspelling.ENABLE
            and spelling_inconsistency == SpellingInconsistency.MISSPELLING.value
        ):
            is_expansion = True
        if is_expansion:
            if flg_input.unify_level in (UnifyLevel.LEXEME, UnifyLevel.WORD_FORM):
                return synonym_group
            elif flg_input.unify_level == UnifyLevel.ABBREVIATION:
                filtered_synonym_group = [s for s in synonym_group if s.abbreviation == abbreviation]
        return filtered_synonym_group

    def normalize_word_by_custom_synonyms(self, word: str) -> Optional[str]:
        """
        Normalize a word by custom synonyms（カスタム同義語で単語を正規化する）

        Args:
            word: Word to normalize（正規化する単語）

        Returns:
            Normalized word（正規化された単語）

        Example:
            normalized_word = normalize_word_by_custom_synonyms(word)
        """
        for k, v in self.custom_synonyms.items():
            if word in v:
                return k
        return None

    def get_morphemes(self, text: str) -> List[str]:
        """
        Get morphemes from text（テキストから形態素を取得）

        Args:
            text: Text to tokenize（トークン化するテキスト）

        Returns:
            Morphemes（形態素）
        """
        tokens = self.tokenizer_obj.tokenize(text, self.mode)
        return [token for token in tokens]

    def get_synonym_group(self, morpheme: Morpheme, is_yougen: bool, is_taigen: bool) -> Optional[List[Synonym]]:
        """
        Get the synonym group of the morpheme（形態素の同義語グループを取得）

        Args:
            morpheme: Morpheme information（形態素情報）

        Returns:
            Synonym group（同義語グループ）

        Example:
            synonym_group = get_synonym_group(morpheme)
        """

        synonym_group_ids = morpheme.synonym_group_ids()
        if synonym_group_ids:
            # Only when there is one synonym group ID. If there are multiple, we cannot determine, so return None. If there is no ID, also return None.
            if len(synonym_group_ids) == 1:
                synonym_group = self.synonyms[synonym_group_ids[0]]
                if is_yougen:
                    return [s for s in synonym_group if s.taigen_or_yougen == TaigenOrYougen.YOUGEN.value]
                if is_taigen:
                    return [s for s in synonym_group if s.taigen_or_yougen == TaigenOrYougen.TAIGEN.value]
        return None

    def get_synonym_value_from_morpheme(
        self, morpheme: Morpheme, synonym_group: List[Synonym], synonym_attr: SynonymField
    ) -> Union[str, int]:
        return next(
            (getattr(item, synonym_attr.value) for item in synonym_group if item.lemma == morpheme.surface()), None
        )
